#! /usr/bin/python


# Copyright (c) 2008-2011, I. Marti-Vidal <mivan@chalmers.se>

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.



#########################################
# Import the necessary modules:

import pyaudio
import numpy.fft
from numpy import *
import array
import gobject
import matplotlib
matplotlib.use('GTKAgg')
import matplotlib.pyplot as plt




#########################################
# Function to just print some help:

def HELP():
 print "\n\n\n"
 print "        ---------------------------------------------------------------"
 print "            Program  ``PYFRINGE''   by Ivan Marti-Vidal  (Jul 2010).    "
 print "        ---------------------------------------------------------------"


 print " - USAGE: Runme inside \"ipython -pylab\" (i.e., using the \"run\" verb). \n"

 print "1.- Run CONFIG(Rate, NCha, NFra), where:"
 print " - Rate is the recording rate in Hz (e.g. 44100)."
 print " - NCha (multiple of 2) is the size of the correlation window (e.g. 512)."
 print " - NFra is the number of frames (fringes) to average before plotting (e.g. 5).\n"

 print "2.- Run RECEIVER(Freq, BW) to set the \"receiver\" digital filter, where:"
 print " - Freq is the central frequency of the \"receiver\" in Hz."
 print " - BW is its bandwidth in Hz. \n"

 print "3.- Run DC(Nsec) to estimate the DC contribution of the mics."
 print " - Nsec is the number of seconds to record silence (e.g. 2). \n"

 print "4.- Run REC(delwin) to begin signal recording and correlation."
 print " - delwin is the delay window to plot (in sec.). If =0, default window is used.\n"

 print "  Other functions:  HELP() (prints these lines); STOP() (stops recording)."
 print "     PLAYB(p) Plays back in real time (if p=1) or stops playback (if p=0)."





##############################################
# Function to define some parameters:

def CONFIG(Rat,Chun,NInt):
 global FORMAT
 global CHANNELS
 global RATE, RATEf, chunk, chunkf
 global aver1, aver2, ninteg
 global exitloop 
 
 aver1 = 0.0
 aver2 = 0.0
 FORMAT = pyaudio.paFloat32
 CHANNELS = 2
 RATE = Rat
 RATEf = float(RATE)
 chunk = Chun
 chunkf = float(chunk)
 ninteg = NInt

 print " \n - Record mode set to ",Rat, " samples per second."
 print " - MAXIMUM RECEIVER FREQUENCY: ",Rat/2.,"Hz." 
 print " - Correlation window set to ", Chun, " channels."
 print " - Each fringe plotted will be the average of ", NInt, " frames.\n"
 print " BINNING CHANGED!  DON'T FORGET TO RUN RECEIVER(Freq,BW) BEFORE REC()!!\n"





#############################################################
# Function to estimate any possible DC contribution:

def DC(RECORD_SECONDS=1.0,av2=0.0):

 if av2==0.0:

# 1.- Default option: no a priori DC calibration given. 
#     Hence, estimate it from the microphones:

 # Open sound:

  p = pyaudio.PyAudio()

  stream = p.open(format = FORMAT,
                channels = 2,
                rate = RATE,
                input = True,
                frames_per_buffer = chunk)

  datapl=''
  print " \n - Estimating DC contribution of microphones (silence, please!)."

 # Read audio data (binary format!):

  for i in range(0, int(RATEf / chunkf * RECORD_SECONDS)):
   data = stream.read(chunk)
   datapl += data

 # Decode binary into Float32:

  zeroth = array.array("f",datapl)

 # Separate (stereo) channels:

  dataz1=[]
  dataz2=[]
  for j in range(len(zeroth)/2):
   dataz1.append(zeroth[2*j])
   dataz2.append(zeroth[2*j+1])

 # Compute average of each channel:

  aver1 = numpy.average(dataz1)
  aver2 = numpy.average(dataz2)

  print " - done! \n"
  print " - DC offsets: ", aver1, "for left ; ", aver2, " for right.\n"
  print " You can run DC(",aver1,",",aver2,")" 
  print " to reset the calibration with these values."

 # Close sound:

  stream.close()
  p.terminate()

 else:

# 2.- A priori DC calibration given. Just set it:

  aver1 = RECORD_SECONDS
  aver2 = av2
  print "\n - DC offsets set to: ", aver1, "for left ; ", aver2, " for right.\n"





#########################################################
# Function to record and plot fringes in real time:

def RECFunc():

 global exitloop, pla
 global plot1, plot2, fig
 global dfmask, data3,data4,xarr
 global intens,crossprod,stream
 global chini, chend, crosstemp

 while (exitloop):

   crossprod.fill(0.0)
   crosstemp.fill(0.0)

   for h in range(ninteg):

    intens[0] = 0.0; intens[1] = 0.0
   # Record chunk frames (output is in binary!):
    data = stream.read(chunk)
   # Decode from binary to Float32:
    data2= array.array("f",data)
   # Separate (stereo) channels and correct DC:
    intens[0] = 0.0; intens[1] = 0.0

    for j in range(chunk):
     data3[j] = data2[2*j]-aver1
     data4[j] = data2[2*j+1]-aver2
    # Compute the intensity of each channel:
     intens[0] += data3[j]*data3[j]
     intens[1] += data4[j]*data4[j]

   # Compute FFT of each channel (and filter in frequency):
    fftdata3=numpy.fft.fft(data3)[chini:chend]
    fftdata4=numpy.fft.fft(data4)[chini:chend]

###########################################################
   # If playback is enabled, scale channel intensities:
   # (linear interpolation of gains between chunks).
    if pla:
     prevint = intens[2]
     intens[2] = numpy.sqrt(intens[0]/intens[1])
     for j in range(chunk):
      data2[2*j+1] *= (prevint*(chunkf-xarr[j])+intens[2]*xarr[j])/chunkf
   # Playback!
     stream.write(data2.tostring(),chunk)
###########################################################

   # Compute correlation as FFT^(-1) of the product of FFTs (i.e., FX correlator):
    crosstemp = crosstemp + (fftdata3*(fftdata4.conjugate()))

   # Filter the signal in frequency (digital filter) and invert FFT:
   crossprod[chini:chend] = crosstemp
   corrcomp = numpy.fft.fft(crossprod)

   # Find the fringe peak and get the fringe intensity:
   correlplot=corrcomp*(corrcomp.conjugate())
   mmax = numpy.argmax(correlplot)
  # Update plot:
   # 1.- Amplitude:
   plot1.set_ydata(numpy.fft.fftshift(correlplot)/correlplot[mmax])
   # 2.- Phase at maximum:
   phase = numpy.arctan2(corrcomp[mmax].imag,corrcomp[mmax].real)*180./numpy.pi 
   plot2.set_xdata([phase,phase]); plot2.set_ydata([0.,1.])

   fig.canvas.draw()

   yield exitloop

 # If the user finishes recording, the audio is closed:
 stream.close()
 p.terminate()
 print "done! \n"




######################################################
# Function to stop recording (and plotting):
def STOP():
 global p,exitloop
 exitloop = False
 print "\n\n RECORDING IS STOPPED!!!!"
 print "You can run CONFIG(R,N,i), RECEIVER(F,B), AND/OR DC(t)" 
 print "and execute REC() again.\n"





######################################################
# Function to begin recording and correlating:

def REC(delwin=0):

 global p,exitloop, pla
 global plot1, plot2, fig
 global data3,data4,xarr
 global intens,crossprod,stream
 global chini, chend, crosstemp

 # Open audio stream:
 pla = False
 p = pyaudio.PyAudio()

 stream = p.open(format = FORMAT,
                channels = CHANNELS,
                rate = RATE,
                input = True,
                output = True,
                frames_per_buffer = chunk)

 # Initialize some variables:

 data3 = numpy.zeros(chunk,dtype="float32")
 data4 = numpy.copy(data3)
 xarr = numpy.arange(chunk,dtype="float32")

 # Initialize plots:

 deltaT=chunkf/RATEf
 rangx1 = 1000.0*(xarr-chunkf/2.)/RATEf
 rangy1 = numpy.copy(data3)
 fig=plt.figure(1,figsize = (18,10))
 fig.clf()
 corr1 = fig.add_axes([0.09,0.25,0.9,0.65])
 plt.title("AKUSTISCHE   INTERFERENZ",fontsize=30)
 plt.xlabel("Zeitunterschied (msec.)",fontsize=20)
 plt.ylabel("Amplitude der Korrelation (norm.)",fontsize=20)
 plt.xticks(fontsize=20)
 plt.yticks(fontsize=20)
 plot1, = corr1.plot(rangx1,rangy1)
 corr1.set_ylim(-0.02,1.02)
 if delwin != 0:
  corr1.set_xlim(-delwin/0.002,delwin/0.002)
 corr2 = fig.add_axes([0.09,0.07,0.9,0.05])
 plt.xlabel("Phase (deg.)",fontsize=15)
 plt.xticks(fontsize=15)
 plt.setp(corr2.get_yticklabels(),visible=False)

 plot2, = corr2.plot([0.,0.],[0.,1.],'-r',lw=4.0)
 corr2.set_xlim(-180.,180.); corr2.set_ylim(0.,1.);
 if delwin != 0:
  corr1.set_xlim(-delwin/0.002,delwin/0.002)

 fig.canvas.draw()
 plt.show()

# Mask for digital filter:
 crosstemp = numpy.zeros(chend-chini,dtype="complex64")

 # Begin data recording!!!
 print " \n - Recording data. Run STOP() to stop recording "
 print "   or run PLAYB(1 or 0) to start/stop a playback."

 exitloop=True

 intens = numpy.zeros(3,dtype="float32")
 crossprod = numpy.zeros(chunk,dtype="complex64")

# Loop to record, correlate, and plot:
 gobject.idle_add(lambda iter=RECFunc() : iter.next())




#############################################
# Function to start/stop the playback:
 
def PLAYB(playback):
 global pla
 if playback ==1:
  pla = True
 elif playback ==0:
  pla = False




#########################################################
# Function to set the channels of the digital filter:

def RECEIVER(Freq,BW):

 global chini
 global chend

 if Freq+BW/2.0 > RATEf/2.0:
  print "\nMaximum frequency in bandwidth is HIGHER than Nyquist sampling!!"
  print "Decrease central frequency of bandwidth and try again."
  chini = 1
  chend = chunk/2
 else:
  NuMin = float(Freq) - float(BW)/2.0
  NuMax = NuMin + float(BW)

  chini = int(around(NuMin/RATEf*chunk))
  chend = int(around(NuMax/RATEf*chunk))

  print " \n - Setting digital filter at frequency ",Freq, " Hz. " 
  print " - Setting bandwidth of digital filter to ", BW, " Hz. \n"

  print " - Filter channels: ",chini," to ",chend




#################################
# print the help at startup:
HELP()
